import json
import os
from typing import Dict, Optional

import numpy as np
import pandas as pd

from .column import Column, IndexColumn
from .tok.bert_tok import BertTok
from .tok.entity_tok import EntTok
from .tok.id_tok import IdTok
from .tok.tokenizer import ListTokenizer
from UniTok.vocab.depot import VocabDepot
from UniTok.vocab.vocab import Vocab


class UniTok:
    VER = 'v2.2'

    def __init__(self):
        self.cols = dict()  # type: Dict[str, Column]
        self.vocab_depot = VocabDepot()
        self.id_col = None  # type: Optional[Column]
        self.data = None

    def add_col(self, col: Column):
        self.cols[col.name] = col
        self.vocab_depot.append(col)

        if isinstance(col.tok, IdTok):
            if self.id_col:
                print(self.id_col.name, col.name)
                raise ValueError('Only one id column (with IdTok) is required!')
            self.id_col = col

        return self

    def add_index_col(self, name='index'):
        col = IndexColumn(name=name)
        self.cols[col.name] = col
        self.vocab_depot.append(col)

        if self.id_col:
            raise ValueError('Already has id column before adding IndexColumn!')
        self.id_col = col

    def read_file(self, df, sep=None):
        if isinstance(df, str):
            use_cols = list(self.cols.keys())
            df = pd.read_csv(df, sep=sep, usecols=use_cols)
        self.data = df
        return self

    def get_col_data(self, col: Column):
        if isinstance(col, IndexColumn):
            return self.data.index
        return self.data[col.name]

    def analyse(self):
        for name in self.vocab_depot.depot:
            vocab = self.vocab_depot.get_vocab(name)
            vocab.init_frequency()

        print('[ COLUMNS ]')
        for col_name in self.cols:
            col = self.cols[col_name]  # type: Column
            print('[ COL:', col.name, ']')
            col.analyse(self.get_col_data(col))
            print()

        print('[ VOCABS ]')
        for vocab_name in self.vocab_depot.col_map:  # type: Vocab
            vocab = self.vocab_depot.depot[vocab_name]
            print('[ VOC:', vocab.name, 'with ', vocab.get_size(), 'tokens ]')
            print('[ COL:', ', '.join(self.vocab_depot.col_map[vocab_name]), ']')
            frequency_dict = vocab.frequency_analyse()
            print('[ FRQ:', frequency_dict, ']')
            print()
        return self

    def tokenize(self):
        if not self.id_col:
            raise ValueError('Id column (with IdTok) is required')

        for col_name in self.cols:
            print('[ COL:', col_name, ']')
            col = self.cols[col_name]  # type: Column
            col.data = []
            col.tokenize(self.get_col_data(col))
        return self

    def get_tok_path(self, col_name, store_dir):
        return self.cols[col_name].tok.vocab.get_store_path(store_dir)

    def store_data(self, store_dir):
        os.makedirs(store_dir, exist_ok=True)

        vocab_info = dict()
        for vocab_name in self.vocab_depot.depot:
            vocab = self.vocab_depot.depot[vocab_name]
            vocab.save(store_dir)
            vocab_info[vocab_name] = dict(
                size=vocab.get_size(),
                cols=self.vocab_depot.col_map[vocab_name],
            )

        data = dict()
        for col_name in self.cols:
            col = self.cols[col_name]
            data[col_name] = np.array(col.data, dtype=object)
        np.save(os.path.join(store_dir, 'data.npy'), data, allow_pickle=True)

        col_info = dict()
        for col_name in self.cols:
            col = self.cols[col_name]
            col_info[col_name] = dict(
                vocab=col.tok.vocab.name,
            )
            if isinstance(col.tokenizer, ListTokenizer):
                max_length = col.tokenizer.max_length
                if max_length < 1:
                    for ids in col.data:
                        if len(ids) > max_length:
                            max_length = len(ids)
                col_info[col_name].update(dict(
                    max_length=max_length,
                    padding=col.tokenizer.padding,
                ))

        meta_data = dict(
            version=UniTok.VER,
            vocab_info=vocab_info,
            col_info=col_info,
            id_col=self.id_col.name,
        )
        json.dump(meta_data, open(os.path.join(store_dir, 'meta.data.json'), 'w'))
        return self


if __name__ == '__main__':
    df = pd.read_csv(
        filepath_or_buffer='~/Data/MIND/MINDlarge/train/news.tsv',
        sep='\t',
        names=['nid', 'cat', 'subCat', 'title', 'abs', 'url', 'titEnt', 'absEnt'],
        usecols=['nid', 'cat', 'subCat', 'title', 'abs'],
    )

    ut = UniTok()
    id_tok = IdTok(name='news')
    cat_tok = EntTok(name='cat')
    txt_tok = BertTok(name='english', vocab_dir='bert-base-uncased')
    cat_tok.vocab.reserve(100)

    ut.add_col(Column(
        name='nid',
        tokenizer=id_tok.as_sing(),
    )).add_col(Column(
        name='cat',
        tokenizer=cat_tok.as_sing()
    )).add_col(Column(
        name='subCat',
        tokenizer=cat_tok.as_sing(),
    )).add_col(Column(
        name='title',
        tokenizer=txt_tok.as_list(),
    )).add_col(Column(
        name='abs',
        tokenizer=txt_tok.as_list(),
    )).read_file(df).tokenize()
    ut.store_data('MINDlarge_train')



from typing import Optional

from UniTok.vocab.vocab import Vocab


class BaseTok:
    def __init__(self, name: str, vocab: Optional[Vocab] = None, pre_handler=None):
        self.PAD = 0
        self.name = name
        self.vocab = vocab or Vocab(name)
        self.pre_handler = pre_handler

    def t(self, obj) -> [int, list]:
        raise NotImplementedError

    def _t(self, obj):
        ids = self.t(obj)
        if isinstance(ids, list):
            return list(filter(lambda index: index > -1, ids))
        if ids == -1:
            raise ValueError('Single Tokenizer should provide vocab, but -1 is given')
        return ids

    def __call__(self, obj):
        return self._t(obj)

    def as_sing(self):
        from .tokenizer import SingT
        return SingT(self)

    def as_list(self, max_length=0, padding=False, slice_post=False, pad_pre=False):
        from .tokenizer import ListT
        return ListT(self, max_length=max_length, padding=padding, slice_post=slice_post, pad_pre=pad_pre)

    def load_vocab(self, store_dir: str, as_path=False):
        self.vocab.load(store_dir=store_dir, as_path=as_path)
        return self



import collections
import math
import os
from typing import Union, List, Optional

import numpy as np
from smartify import E


@E.register(id_processor=E.idp_cls_prefix())
class VocabError:
    NotEditable = E('Vocab {} is not editable, but new word [{}] appears')
    NotEmptyForReserve = E('Vocab {} is not empty and not allowed reserve operation')


class Vocab:
    __VOCAB_ID = 0

    @classmethod
    def get_vocab_id(cls):
        vocab_id = cls.__VOCAB_ID
        cls.__VOCAB_ID += 1
        return vocab_id

    def __init__(self, name: str):
        if not isinstance(name, str):
            raise ValueError('Vocab name should be string')

        self.name = name
        self.obj2index, self.index2obj = {}, {}
        self.editable = True
        self.frequency_mode = False
        self.oov_default = None
        self.vocab_id = Vocab.get_vocab_id()
        self.frequency = {}
        self.max_frequency = 0

        self.frequent_vocab = []
        self.reserve_tokens = None

    def init_frequency(self):
        self.frequency = {}
        self.max_frequency = 0

    def frequency_count(self, *ids):
        for index in ids:
            if index not in self.frequency:
                self.frequency[index] = 0
            self.frequency[index] += 1
            if self.max_frequency < self.frequency[index]:
                self.max_frequency = self.frequency[index]

    def trim_vocab(self, min_frequency=1, oov_default=None):
        self.oov_default = self.oov_default or oov_default
        self.frequent_vocab = []
        for index in self.frequency:
            if self.frequency[index] >= min_frequency:
                self.frequent_vocab.append(self.index2obj[index])
        self.index2obj = dict()
        self.obj2index = dict()

        if self.reserve_tokens is not None:
            self.reserve(self.reserve_tokens)
        self.extend(self.frequent_vocab)

        self.frequency_mode = True

    def frequency_analyse(self):
        max_count = self.max_frequency
        digits_max = 10
        while digits_max < max_count:
            digits_max = digits_max * 10

        bounds = []
        while digits_max >= 10:
            digits_min = digits_max // 10
            left_bound = (np.arange(9)[::-1] + 1) * digits_min
            right_bound = left_bound + digits_min
            bounds.extend(zip(left_bound, right_bound))
            digits_max = digits_min
        bounds.append((0, 1))
        bounds.reverse()

        bound_dict = dict()
        for bound in bounds:
            bound_dict[bound] = 0

        for index in self.frequency:
            for bound in bounds:
                if bound[1] > self.frequency[index] >= bound[0]:
                    bound_dict[bound] += 1
                    break

        for bound in bounds:
            if not bound_dict[bound]:
                del bound_dict[bound]

        return bound_dict

    def extend(self, objs):
        for obj in objs:
            self.append(obj)
        return self

    def append(self, obj) -> int:
        if obj not in self.obj2index:
            if self.frequency_mode:
                if self.oov_default is not None:
                    return self.oov_default
                return -1
            if not self.editable:
                if self.oov_default is not None:
                    return self.oov_default
                raise VocabError.NotEditable(self.name, obj)
            index = len(self.index2obj)
            self.obj2index[obj] = index
            self.index2obj[index] = obj
        return self.obj2index[obj]

    def reserve(self, tokens: Union[int, List[any]]):
        if self.get_size():
            raise VocabError.NotEmptyForReserve(self.name)

        self.reserve_tokens = tokens
        if isinstance(tokens, int):
            digits = int(math.log10(tokens))
            token_template = '[UNUSED%%0%sd]' % digits
            for token in range(tokens):
                self.append(token_template % token)
        else:
            for token in tokens:
                self.append(token)
        return self

    def get_tokens(self):
        return [self.index2obj[i] for i in range(len(self.index2obj))]

    def allow_edit(self):
        self.editable = True
        return self

    def deny_edit(self):
        self.editable = False
        return self

    def get_store_path(self, store_dir):
        return os.path.join(store_dir, 'tok.{}.dat'.format(self.name))

    def load(self, store_dir: str, as_path=False):
        store_path = store_dir if as_path else self.get_store_path(store_dir)

        self.obj2index, self.index2obj = {}, {}
        with open(store_path, 'r') as f:
            objs = f.read().split('\n')[:-1]
        for index, obj in enumerate(objs):
            self.obj2index[obj] = index
            self.index2obj[index] = obj

        return self

    def save(self, store_dir):
        store_path = self.get_store_path(store_dir)
        with open(store_path, 'w') as f:
            for i in range(len(self.index2obj)):
                f.write('{}\n'.format(self.index2obj[i]))
        return self

    def get_size(self):
        return len(self.index2obj)



from tqdm import tqdm

from .analysis.length_analysis import LengthAnalysis
from .tok import IdTok
from .tok.tokenizer import Tokenizer, ListTokenizer


class Column:
    def __init__(self, name, tokenizer: Tokenizer):
        self.name = name
        self.tokenizer = tokenizer
        self.tok = tokenizer.tok
        self.data = []
        self.length_analysis = LengthAnalysis()

    def tokenize(self, objs):
        self.data = []
        for obj in tqdm(objs):
            ids = self.tokenizer.tokenize(obj)
            self.data.append(ids)

    def analyse(self, objs):
        if isinstance(self.tokenizer, ListTokenizer):
            self.length_analysis.clean()
            for obj in tqdm(objs):
                ids = self.tok(obj)
                self.tok.vocab.frequency_count(*ids)
                self.length_analysis.push(len(ids))
            self.length_analysis.analyse()
        else:
            print('[NOT ListTokenizer]')
            for obj in tqdm(objs):
                self.tok.vocab.frequency_count(self.tok(obj))


class IndexColumn(Column):
    def __init__(self, name='index'):
        super().__init__(name, tokenizer=IdTok(name=name).as_sing())



import abc

from .tok import BaseTok


class Tokenizer(abc.ABC):
    def __init__(self, tok: BaseTok):
        self.tok = tok

    def tokenize(self, obj):
        raise NotImplementedError


class SingTokenizer(Tokenizer):
    def tokenize(self, obj):
        return self.tok(obj)


class ListTokenizer(Tokenizer):
    def __init__(self,
                 tok: BaseTok,
                 max_length: int = 0,
                 padding: bool = False,
                 slice_post: bool = False,
                 pad_pre: bool = False,
                 ):
        super(ListTokenizer, self).__init__(tok=tok)
        self.max_length = max_length
        self.slice_post = slice_post
        self.padding = padding
        self.pad_pre = pad_pre

    def tokenize(self, obj):
        ids = self.tok(obj)  # type: list
        if self.max_length > 0:
            ids = ids[-self.max_length:] if self.slice_post else ids[:self.max_length]
            if self.padding:
                pads = [self.tok.PAD] * (self.max_length - len(ids))
                if self.pad_pre:
                    ids = pads + ids
                else:
                    ids.extend(pads)
        return ids


T = Tokenizer
SingT = SingTokenizer
ListT = ListTokenizer



import collections
import math

import termplot


class Plot:
    def __init__(self,
                 array: list,
                 groups: int = 50,
                 height: int = 10,
                 ):
        self.array = array
        self.min = min(self.array)
        self.max = max(self.array)

        self.height = height
        self.groups = groups
        self.range = self.max - self.min + 1
        if self.max - self.min + 1 < self.groups:
            self.groups = self.range
        self.interval = math.ceil(self.range / self.groups)
        self.groups = math.ceil(self.range / self.interval)

        self.counts = [0] * self.groups
        for n in self.array:
            i = (n - self.min) // self.interval
            self.counts[i] += 1

    def plot(self):
        print('[ X-INT:', self.interval, ']')
        print('[ Y-INT:', max(self.counts) // self.height, ']')
        termplot.plot(
            x=self.counts,
            plot_char='|',
            plot_height=self.height,
        )
        print('-' * self.groups)


if __name__ == '__main__':
    a = [1, 2, 3, 4, -5, 5, -4, -1, 0, -10, -4, -2, 3, 5, 8, 10, 12, 10, 8, 7, 6, 5, 4, 3, 2, 1]
    print(collections.Counter(a))
    Plot(a).plot()



import pandas as pd

from .tok import BaseTok
from UniTok.vocab.vocab import Vocab


class SplitTok(BaseTok):
    def __init__(self, name, sep, **kwargs):
        super(SplitTok, self).__init__(name=name, **kwargs)
        self.sep = sep

    def t(self, obj,):
        ids = []
        if pd.notnull(obj):
            ts = obj.split(self.sep)
            for t in ts:
                if self.pre_handler:
                    t = self.pre_handler(t)
                ids.append(self.vocab.append(t))
        return ids



from .unitok import UniTok
from UniTok.vocab.vocab import Vocab, VocabError
from UniTok.vocab.depot import VocabDepot, VocabDepotError
from .column import Column
from .analysis import LengthAnalysis, Plot
from .unidep import UniDep

__all__ = [
    UniTok,
    UniDep,
    LengthAnalysis,
    Plot,
    Column,
    column,
    analysis,
    Vocab, VocabError,
    VocabDepot, VocabDepotError,
]



from .tok import BaseTok
from UniTok.vocab.vocab import Vocab


class EntTok(BaseTok):
    def t(self, obj):
        if self.pre_handler:
            obj = self.pre_handler(obj)
        return self.vocab.append(str(obj))



import json
import os
import random
from typing import Dict, List

import numpy as np

from .unitok import UniTok
from .classify import Classify
from .vocab import VocabDepot, Vocab


class UniDep:
    def __init__(self, store_dir):
        self.store_dir = os.path.expanduser(store_dir)

        self.meta_path = os.path.join(self.store_dir, 'meta.data.json')
        self.meta_data = Classify(json.load(open(self.meta_path)))

        if self.meta_data.version != UniTok.VER:
            print('UniTok version not match, it may occur unexpected errors in loading phase!')

        self.data_path = os.path.join(self.store_dir, 'data.npy')
        self.data = np.load(self.data_path, allow_pickle=True)
        try:
            self.data = self.data.item()  # type: dict
        except Exception as err:
            print(err)

        self.id_col = self.meta_data.id_col
        self.id_vocab = self.meta_data.col_info.d[self.id_col].vocab
        self.sample_size = self.meta_data.vocab_info.d[self.id_vocab].size
        print('Loaded', self.sample_size, 'samples!')

        data_sample_size = len(self.data[self.id_col])
        if self.sample_size != data_sample_size:
            print('Resize sample size to', data_sample_size)
            self.sample_size = data_sample_size

        self.col_info = self.meta_data.col_info
        self.vocab_info = self.meta_data.vocab_info

        self.vocab_depot = VocabDepot()
        for vocab_name in self.vocab_info.d:
            self.vocab_depot.append(Vocab(name=vocab_name).load(self.store_dir))
        self.id2index = self.vocab_depot.depot[self.id_vocab].obj2index

        self.index_order = list(range(self.sample_size))
        self.union_depots = dict()  # type: Dict[str, List[UniDep]]

    @staticmethod
    def merge_col(c1: Classify, c2: Classify):
        for col_name in c2.d:
            if col_name in c1.d and c1.d[col_name].dict() != c2.d[col_name].dict():
                raise ValueError('Column Config Conflict In Key {}'.format(col_name))
        d = c1.dict()
        d.update(c2.dict())
        return Classify(d)

    @staticmethod
    def merge_vocab(c1: Classify, c2: Classify):
        for vocab_name in c2.d:
            if vocab_name in c1.d and c1.d[vocab_name].size != c2.d[vocab_name].size:
                raise ValueError('Vocab Config Conflict In Key {}'.format(vocab_name))
        d = c1.dict()
        d.update(c2.dict())
        return Classify(d)

    def union(self, *depots: 'UniDep'):
        for depot in depots:
            if depot.id_col not in self.col_info.d:
                raise ValueError('Current Depot Has No Column Named {}'.format(depot.id_col))

            if depot.id_col not in self.union_depots:
                self.union_depots[depot.id_col] = []
            self.union_depots[depot.id_col].append(depot)
            self.col_info = self.merge_col(self.col_info, depot.col_info)
            self.vocab_info = self.merge_vocab(self.vocab_info, depot.vocab_info)
            self.meta_data.col_info = self.col_info
            self.meta_data.vocab_info = self.vocab_info

    def is_list_col(self, col_name):
        return 'max_length' in self.col_info.d[col_name].d

    def get_vocab_size(self, col_name, as_vocab=False):
        vocab_id = col_name if as_vocab else self.get_vocab(col_name)
        return self.vocab_info.d[vocab_id].size

    def get_vocab(self, col_name):
        return self.col_info.d[col_name].vocab

    def get_max_length(self, col_name):
        if self.is_list_col(col_name):
            return self.col_info.d[col_name].max_length

    def get_sample_by_id(self, obj_id):
        return self.pack_sample(self.id2index[obj_id])

    def shuffle(self, shuffle=True):
        if shuffle:
            random.shuffle(self.index_order)
        else:
            self.index_order = list(range(self.sample_size))

    def pack_sample(self, index):
        sample = dict()
        for col_name in self.col_info.d:
            if col_name in self.data:
                sample[col_name] = self.data[col_name][index]
                if col_name in self.union_depots:
                    for depot in self.union_depots[col_name]:
                        sample.update(depot[sample[col_name]])
        return sample

    def __getitem__(self, index):
        index = self.index_order[index]
        return self.pack_sample(index)

    def __len__(self):
        return self.sample_size



from typing import Dict

from smartify import E

from .vocab import Vocab


@E.register(id_processor=E.idp_cls_prefix())
class VocabDepotError:
    ConflictName = E('Conflict name: {}')
    NotFound = E('Not found vocab {}')


class VocabDepot:
    def __init__(self):
        self.depot = {}  # type: Dict[str, Vocab]
        self.col_map = {}

    def append(self, col_or_vocab):
        if isinstance(col_or_vocab, Vocab):
            vocab = col_or_vocab
        else:
            col = col_or_vocab
            vocab = col_or_vocab.tok.vocab
            if vocab.name in self.col_map:
                self.col_map[vocab.name].append(col.name)
            else:
                self.col_map[vocab.name] = [col.name]

        assert vocab.name is not None
        if vocab.name in self.depot and self.depot[vocab.name] != vocab:
            raise VocabDepotError.ConflictName(vocab.name)
        self.depot[vocab.name] = vocab

    def get_vocab(self, name) -> Vocab:
        if name in self.depot:
            return self.depot[name]
        raise VocabDepotError.NotFound(name)

    def __call__(self, name) -> Vocab:
        return self.get_vocab(name)



from .plot import Plot


class LengthAnalysis:
    def __init__(self):
        self.lengths = []
        self.max_length = -1

    def push(self, length):
        self.lengths.append(length)
        if length > self.max_length:
            self.max_length = length

    def analyse(self):
        max_length = max(self.lengths)
        min_length = min(self.lengths)
        avg_length = '%.4f' % (sum(self.lengths) * 1.0 / len(self.lengths))
        print('[ MIN:', min_length, ']')
        print('[ MAX:', max_length, ']')
        print('[ AVG:', avg_length, ']')
        Plot(self.lengths, groups=100).plot()

    def clean(self):
        self.lengths = []
        self.max_length = -1


if __name__ == '__main__':
    import random

    analysis = LengthAnalysis()
    analysis.clean()
    for _ in range(67923):
        analysis.push(random.randint(0, 99999) % 100)
    analysis.analyse()



import pandas as pd
from transformers import BertTokenizer

from .tok import BaseTok


class BertTok(BaseTok):
    def __init__(self, name, vocab_dir):
        super(BertTok, self).__init__(name=name)
        self.tokenizer = BertTokenizer.from_pretrained(vocab_dir)
        self.vocab.extend(self.tokenizer.vocab)

    def t(self, obj) -> [int, list]:
        if pd.notnull(obj):
            ts = self.tokenizer.tokenize(obj)
            ids = self.tokenizer.convert_tokens_to_ids(ts)
        else:
            ids = []
        return ids

    def get_vocabs(self):
        return self.tokenizer.vocab



from .bert_tok import BertTok
from .entity_tok import EntTok
from .id_tok import IdTok
from .split_tok import SplitTok
from .tok import BaseTok
from .tokenizer import Tokenizer, SingTokenizer, ListTokenizer, T, SingT, ListT

__all__ = [
    BaseTok, BertTok, EntTok, IdTok, SplitTok,
    Tokenizer, SingTokenizer, ListTokenizer, T, SingT, ListT,
]



class Classify:
    def dict_list(self, l: list):
        new_l = []
        for v in l:
            if isinstance(v, Classify):
                v = v.dict()
            elif isinstance(v, list):
                v = self.dict_list(v)
            new_l.append(v)
        return new_l

    def dict(self):
        d = dict()
        for k in self.d:
            if isinstance(self.d[k], list):
                d[k] = self.dict_list(self.d[k])
            elif isinstance(self.d[k], Classify):
                d[k] = self.d[k].dict()
            else:
                d[k] = self.d[k]
        return d

    def iter_list(self, l: list):
        new_l = []
        for v in l:
            if isinstance(v, dict):
                v = Classify(v)
            elif isinstance(v, list):
                v = self.iter_list(v)
            new_l.append(v)
        return new_l

    def iter_dict(self, d: dict):
        for k in d:
            if isinstance(d[k], dict):
                d[k] = Classify(d[k])
            elif isinstance(d[k], list):
                d[k] = self.iter_list(d[k])
        return d

    def __init__(self, d: dict):
        self.d = self.iter_dict(d)

    def __getattr__(self, item):
        return self.d[item]

    def __setattr__(self, key, value):
        if key in ['d']:
            object.__setattr__(self, key, value)
        else:
            if isinstance(value, dict):
                value = Classify(value)
            self.d[key] = value


if __name__ == '__main__':
    c = Classify(dict(a=1, b=dict(x='a', y=[3, [4, dict(y='l')], dict(z=1)])))
    print(c.b.y[1][1].y)



from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='UniTok',
    version='2.2.2.5',
    keywords=['token', 'tokenizer', 'bert'],
    description='Unified Tokenizer',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT Licence',
    url='https://github.com/Jyonn/UnifiedTokenizer',
    author='Jyonn Liu',
    author_email='i@6-79.cn',
    platforms='any',
    packages=find_packages(),
    install_requires=[
        'smartify==0.0.2',
        'termplot==0.0.2',
        'tqdm',
        'numpy',
        'pandas'
    ],
)
